/*
   Copyright (C) 1998  Ulric Eriksson <ulric@edu.stockholm.se>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the Licence, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.
 */

/* 
 * Application.c - Application composite widget
 */

#include	<X11/IntrinsicP.h>
#include	<X11/StringDefs.h>
#include	<X11/Xmu/Misc.h>
#include	<X11/Xaw/Command.h>
#include	"ApplicationP.h"

/****************************************************************
 *
 * Application Resources
 *
 ****************************************************************/

/* These are bogus; they are hand-me-downs from Box */

static XtResource resources[] = {
    { XtNhSpace, XtCHSpace, XtRDimension, sizeof(Dimension),
		XtOffsetOf(ApplicationRec, application.h_space),
		XtRImmediate, (XtPointer)4 },
    { XtNvSpace, XtCVSpace, XtRDimension, sizeof(Dimension),
		XtOffsetOf(ApplicationRec, application.v_space),
		XtRImmediate, (XtPointer)4 }
};

/****************************************************************
 *
 * Full class record constant
 *
 ****************************************************************/

#if 0
static void Initialize();
#endif
static void Resize();
static Boolean SetValues();
static XtGeometryResult GeometryManager();
static void ChangeManaged();

ApplicationClassRec applicationClassRec = {
  {
/* core_class fields      */
    /* superclass         */    (WidgetClass) &compositeClassRec,
    /* class_name         */    "Application",
    /* widget_size        */    sizeof(ApplicationRec),
    /* class_initialize   */    NULL,
    /* class_part_init    */	NULL,
    /* class_inited       */	FALSE,
    /* initialize         */    NULL,
    /* initialize_hook    */	NULL,
    /* realize            */    XtInheritRealize,
    /* actions            */    NULL,
    /* num_actions	  */	0,
    /* resources          */    resources,
    /* num_resources      */    XtNumber(resources),
    /* xrm_class          */    NULLQUARK,
    /* compress_motion	  */	TRUE,
    /* compress_exposure  */	TRUE,
    /* compress_enterleave*/	TRUE,
    /* visible_interest   */    FALSE,
    /* destroy            */    NULL,
    /* resize             */    Resize,
    /* expose             */    NULL,
    /* set_values         */    SetValues,
    /* set_values_hook    */	NULL,
    /* set_values_almost  */    XtInheritSetValuesAlmost,
    /* get_values_hook    */	NULL,
    /* accept_focus       */    NULL,
    /* version            */	XtVersion,
    /* callback_private   */    NULL,
    /* tm_table           */    NULL,
    /* query_geometry     */	XtInheritQueryGeometry,
    /* display_accelerator*/	XtInheritDisplayAccelerator,
    /* extension          */	NULL
  },{
/* composite_class fields */
    /* geometry_manager   */    GeometryManager,
    /* change_managed     */    ChangeManaged,
    /* insert_child	  */	XtInheritInsertChild,
    /* delete_child	  */	XtInheritDeleteChild,
    /* extension          */	NULL
  },{
/* Application class fields */
    /* empty		  */	0
  }
};

WidgetClass applicationWidgetClass = (WidgetClass)&applicationClassRec;


/****************************************************************
 *
 * Private Routines
 *
 ****************************************************************/

/*
 * Do a layout, actually assigning positions.
 *
 * For n child widgets numbered 0 to n-1, lay them out like this:
 *
 * 0 - n-3	Each 30 high, at the top
 * n-1		30 high, at the bottom
 * n-2		All the remaining space, bu no less than 2
 *
 * Each child widget is as wide as the ApplicationWidget.
 */

static void DoLayout(ApplicationWidget sw)
{
	Dimension w3, h1, h3;
	int n = sw->composite.num_children;
	int i;

	w3 = sw->core.width;
	h3 = sw->core.height;
	h1 = 0;

	for (i = 0; i < n-2; i++) {
		XtConfigureWidget(sw->composite.children[i],
				0, h1, w3, 30, 0);
		h1 += 30;
	}
	/* there are three possible situations when we get here:
	 * 1. There are no widgets left (i >= n). Do nothing.
	 * 2. There is one widget left (i == n-1). Use all space for it.
	 * 3. There are two widgets left (i < n-1).
	 */
	if (i == n-1) {
		XtConfigureWidget(sw->composite.children[i],
				0, h1, w3, h3-h1, 0);

		return;
	} else if (i == n-2) {
		XtConfigureWidget(sw->composite.children[i],
			0, h1, w3, h3-h1-30, 0);
		XtConfigureWidget(sw->composite.children[i+1],
			0, h3-30, w3, 30, 0);
	}
}

/*
 * Actually layout application
 */

static void Resize(Widget w)
{
	DoLayout((ApplicationWidget)w);
} /* Resize */

/*
 * Geometry Manager
 *
 * 'reply' is unused; we say only yeay or nay, never almost.
 */

static XtGeometryResult GeometryManager(Widget w,
		XtWidgetGeometry request, XtWidgetGeometry reply)
{
	return XtGeometryYes;
}

static void ChangeManaged(Widget w)
{
    DoLayout((ApplicationWidget)w);
}

static Boolean SetValues(Widget current, Widget request, Widget new,
		ArgList args, Cardinal *num_args)
{
    return False;
}
